'''
fitdistcp package. Visit fitdistcp.info for more information.
Provides modules 
    - genextreme: generalised extreme value distribution
    - genextreme_p1: generalised extreme value distribution, with one predictor
    - genpareto: generalised pareto distribution
'''

__all__ = [
    'genextreme', 
    'genextreme_p1', 
    'genpareto',
    'expon',
	'gamma',
	'gumbel',
	'lnorm',
	'norm',
	'weibull',
    'genextreme_p12',
    'reltest'
]